/*****************************************************************************
 *
 * FILE:	param.c
 * DESCRIPTION:	Cockatrice: parameter manipulator
 * DATE:	Sat, Mar 16 2002
 * UPDATED:	Sun, Aug 19 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: param.c,v 1.3 2007/08/19 16:48:36 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "cockatrice.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	PARAMETER_TABLE_SIZE	(13)
typedef struct _param {
  char *		name;
  VList *		vlist;	/* value list */
  struct _param *	next;
} Param;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static Param *		lookup(const char *);
static unsigned int	hash(const char *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Param *	param[PARAMETER_TABLE_SIZE];

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	cc_param_init
 * DESCRIPTION:		initialize a parameter table
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
cc_param_init(void)
{
  register int	i;

  for (i = 0; i < PARAMETER_TABLE_SIZE; i++) {
    param[i] = NULL;
  }
}

/*
 * FUNCTION NAME:	cc_param_finish
 * DESCRIPTION:		release memory area of parameter table
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
cc_param_finish(void)
{
  register int	i;

  for (i = 0; i < PARAMETER_TABLE_SIZE; i++) {
    register Param *	pp;
    register Param *	pp_next;

    if (param[i] == NULL) {
      continue;
    }

    for (pp = param[i]; pp != NULL; pp = pp_next) {
      pp_next = pp->next;
      cc_free(pp->name);
      if (pp->vlist != NULL) {
	register VList *	vp;
	register VList *	vp_next;

	for (vp = pp->vlist; vp != NULL; vp = vp_next) {
	  vp_next = vp->next;
	  cc_free(vp->value);
	  free(vp);
	}
	pp->vlist = NULL;
      }
      free(pp);
    }

    param[i] = NULL;
  }
}

/*
 * FUNCTION NAME:	cc_param_put
 * DESCRIPTION:		store CGI parameter into the table
 * ARGUMENTS:		name, value
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
int
cc_param_put(name, value)
	const char *	name;
	const char *	value;
{
  VList *	vp;	/* value list pointer */
  Param *	pp;	/* new parameter pointer */
  unsigned int	h;	/* hash value */

  if (name == NULL) {
    cgi_errno = CGIERR_INVALID_ARGUMENTS;
    return -1;
  }

  vp = (VList *)cc_calloc(1, sizeof(VList));
  if (vp == NULL) {
    return -1;
  }
  vp->value = (value != NULL && value[0] != EOL) ? cc_strdup(value) : NULL;
  vp->next  = NULL;

  pp = lookup(name);
  if (pp == NULL) {	/* first register */
    pp = (Param *)cc_calloc(1, sizeof(Param));
    if (pp == NULL) {
      return -1;
    }

    h		= hash(name);
    pp->name	= cc_strdup(name);
    pp->vlist	= vp;
    pp->next	= param[h];
    param[h]	= pp;
  }
  else {		/* already registered name */
    vp->next	= pp->vlist;
    pp->vlist	= vp;
  }

  return 0;
}

/*
 * FUNCTION NAME:	cc_param_get
 * DESCRIPTION:		get CGI value from the table
 * ARGUMENTS:		name
 * RETURN VALUE:	value
 */
const char *
cc_param_get(name)
	const char *	name;
{
  register Param *	pp;

  pp = lookup(name);
  if (pp != NULL) {
    return pp->vlist->value;
  }

  return NULL;
}

/*
 * FUNCTION NAME:	cc_params_get
 * DESCRIPTION:		get CGI values from the table
 * ARGUMENTS:		name
 * RETURN VALUE:	values
 */
VList *
cc_params_get(name)
	const char *	name;
{
  register Param *	pp;

  pp = lookup(name);
  if (pp != NULL) {
    return pp->vlist;
  }

  return NULL;
}

/*
 * FUNCTION NAME:	lookup
 * DESCRIPTION:		look up a parameter name
 * ARGUMENTS:		name
 * RETURN VALUE:	Successful: parameter pointer, otherwise: NULL
 */
static Param *
lookup(name)
	const char *	name;
{
  if (name != NULL) {
    unsigned int	h;	/* hash value */
    register Param *	pp;	/* working pointer */

    h = hash(name);
    for (pp = param[h]; pp != NULL; pp = pp->next) {
      if (pp->name != NULL && strcmp(name, pp->name) == 0) {
	return pp;
      }
    }
  }

  return NULL;
}

/*
 * FUNCTION NAME:	hash
 * DESCRIPTION:		calculates a hash value of string
 * ARGUMENTS:		string
 * RETURN VALUE:	hash value
 */
static unsigned int
hash(str)
	const char *	str;
{
  register char *	p;
  register unsigned int	h;

  p = (char *)str;
  for (h = *p++; *p != EOL; p++) {
    h += *p;
  }

  return h % PARAMETER_TABLE_SIZE;
}
