/*****************************************************************************
 *
 * FILE:	upload.c
 * DESCRIPTION:	Cockatrice: upload manipulator
 * DATE:	Tue, Mar 19 2002
 * UPDATED:	Sun, Aug 19 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: upload.c,v 1.4 2007/08/19 16:48:36 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif	/* HAVE_SYS_TYPES_H */
#include <md5.h>
#include <sha.h>
#include <sha256.h>
#include <ripemd.h>
#include "cockatrice.h"
#include "module.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	CONTENT_DISPOSITION		"Content-Disposition"
#define	CONTENT_TYPE			"Content-Type"
#define	CONTENT_TRANSFER_ENCODING	"Content-Transfer-Encoding"

typedef struct {	/* RFC1867 */
  struct {	/* Content-Disposition ... RFC1806 */
    char *	type;	/* inline/attachment/form-data */
    char *	name;
    char *	filename;
  } disposition;
  struct {	/* Content-Type: type/subtype; parameter */
    char *	mediaType;
    char *	parameter;
  } type;
  /* 7bit/8bit/binary/quoted-printable/base64/ietf-token/x-token */
  char *	transferEncoding;
} ContentHeader;

typedef struct _ListManager {
  char *		name;
  FILE *		fp;
  struct _ListManager *	next;
} ListManager;

#ifndef	MD5_DIGEST_LENGTH
#define	MD5_DIGEST_LENGTH	(16)
#endif	/* MD5_DIGEST_LENGTH */
#ifndef	SHA_DIGEST_LENGTH
#define	SHA_DIGEST_LENGTH	(20)
#endif	/* SHA_DIGEST_LENGTH */
#ifndef	SHA256_DIGEST_LENGTH
#define	SHA256_DIGEST_LENGTH	(32)
#endif	/* SHA256_DIGEST_LENGTH */
#ifndef	RIPEMD160_DIGEST_LENGTH
#define	RIPEMD160_DIGEST_LENGTH	(20)
#endif	/* RIPEMD160_DIGEST_LENGTH */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static FILE *	cm_upload_open(const char *);
static int	cm_upload_store(const char *, const char *);
static char *	cm_upload_digest(const char *, digest_t);

static char *	get_separator(const char *);
static int	read_header(ContentHeader *);
static char *	get_data(const char *);
static char *	get_value(char *);
static int	read_entity(const char *, const char *);
static char *	get_entity(const char *);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static ListManager *	files = NULL;	/* upload file list */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
Content-type: multipart/form-data; boundary=------------------------------29531351678801353387200
*/

/*
 * FUNCTION NAME:	cc_upload_init
 * DESCRIPTION:		parse multipart form data
 * ARGUMENTS:		none
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
int
cc_upload_init(void)
{
  char *	contentType = getenv("CONTENT_TYPE");

  if (contentType != NULL && strstr(contentType, MULTIPART_FORM) != NULL) {
    char *		separator;
    ContentHeader	content;

    separator = get_separator(contentType);
    if (separator == NULL) {
      cgi_errno = CGIERR_ILLEGAL_MULTIPART_FORM;
      return -1;
    }

    content.disposition.type	 = NULL;
    content.disposition.name	 = NULL;
    content.disposition.filename = NULL;
    content.type.mediaType	 = NULL;
    content.type.parameter	 = NULL;
    content.transferEncoding	 = NULL;

    while (!feof(stdin)) {
      if (read_header(&content) == -1) {
	return -1;
      }

      if (content.disposition.name != NULL) {
	if (content.disposition.filename != NULL) {
	  cc_param_put(content.disposition.name, content.disposition.filename);
	  read_entity(separator, content.disposition.name);
	}
	else {
	  char *	value = get_entity(separator);

	  cc_param_put(content.disposition.name, value);
	}
      }
      cc_free(content.disposition.type);
      cc_free(content.disposition.name);
      cc_free(content.disposition.filename);
      cc_free(content.type.mediaType);
      cc_free(content.type.parameter);
      cc_free(content.transferEncoding);
    }

    return 0;
  }

  return -1;
}

/*
 * FUNCTION NAME:	cc_upload_finish
 * DESCRIPTION:		close all upload files
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
void
cc_upload_finish(void)
{
  register ListManager *	lmp;
  register ListManager *	lmp_next;

  for (lmp = files; lmp != NULL; lmp = lmp_next) {
    lmp_next = lmp->next;
    cc_free(lmp->name);
    if (lmp->fp != NULL) {
      fclose(lmp->fp);
    }
    free(lmp);
  }
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	get_separator
 * DESCRIPTION:		get a separator of multipart form data
 * ARGUMENTS:		Content-Type line
 * RETURN VALUE:	Successful: separator, otherwise: NULL
 */
static char *
get_separator(line)
	const char *	line;
{
  char *	separator = NULL;

  if (line != NULL) {
    char *	p;

    p = (char *)strstr(line, "boundary=");
    if (p != NULL) {
      (void)asprintf(&separator, "--%s", p + 9);
    }
  }

  return separator;
}

/*
 * FUNCTION NAME:	read_header
 * DESCRIPTION:		read a header of multipart form data
 * ARGUMENTS:		header received buffer
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
read_header(content)
	ContentHeader *	content;
{
  char	buf[BUFSIZ];

  memset(buf, 0, sizeof(buf));
  while (!feof(stdin)) {
    if (fgets(buf, sizeof(buf), stdin) == NULL) {
      return 0;
    }
    if (buf[0] == CR && buf[1] == LF) {
      return 0;
    }
    else if (strncasecmp(buf, CONTENT_DISPOSITION, 19) == 0) {
#define	MAX_DISP_ARGS		(3)
      char *	args[MAX_DISP_ARGS];
      char **	ap;
      char *	disposition = get_data(buf);
      int	i;

      if (disposition == NULL) {
	return -1;
      }

#define	DISP_ARG_TYPE		(0)
#define	DISP_ARG_NAME		(1)
#define	DISP_ARG_FILENAME	(2)
      for (i = 0; i < MAX_DISP_ARGS; i++) {
	args[i] = NULL;
      }
      for (i = 0, ap = args; (*ap = strsep(&disposition, ";")) != NULL; i++) {
	args[i] = *ap;
	if (**ap != EOL) {
	  if (++ap >= &args[MAX_DISP_ARGS]) {
	    break;	/* error: arguments overflow */
	  }
	}
      }
      if (args[DISP_ARG_TYPE] != NULL) {
	content->disposition.type = get_value(args[DISP_ARG_TYPE]);
      }
      if (args[DISP_ARG_NAME] != NULL) {
	content->disposition.name = get_value(args[DISP_ARG_NAME]);
      }
      if (args[DISP_ARG_FILENAME] != NULL) {
	content->disposition.filename = get_value(args[DISP_ARG_FILENAME]);
      }
      free(disposition);
    }
    else if (strncasecmp(buf, CONTENT_TYPE, 12) == 0) {
      char *	type = get_data(buf);
      char *	p;

      if (type == NULL) {
	return -1;
      }

      p = strchr(type, ';');
      if (p != NULL) {
	*p = EOL;
	for (++p; *p == SPC; p++) {	/* remove white space */
	  ;	/* nothing */
	}
	content->type.parameter = cc_strdup(p);
      }
      content->type.mediaType = cc_strdup(type);
      free(type);
    }
    else if (strncasecmp(buf, CONTENT_TRANSFER_ENCODING, 25) == 0) {
      content->transferEncoding = get_data(buf);
    }
    else {
      /* ignored other header fields */
    }
    memset(buf, 0, sizeof(buf));
  }

  return -1;
}

/*
 * FUNCTION NAME:	get_data
 * DESCRIPTION:		get data of content header field
 * ARGUMENTS:		content header field
 * RETURN VALUE:	Successful: data, otherwise: NULL
 */
static char *
get_data(field)
	const char *	field;
{
  register char *	p;

  p = strchr(field, ':');
  if (p != NULL) {
    for (++p; *p == SPC; p++) {	/* remove white space */
      ;	/* nothing to do */
    }
    return cc_strdup(p);
  }

  return NULL;
}

/*
 * FUNCTION NAME:	get_value
 * DESCRIPTION:		get a value of data field
 * ARGUMENTS:		data field
 * RETURN VALUE:	Successful: value, otherwise: NULL
 */
static char *
get_value(field)
	char *	field;
{
  register char *	p;

  p = strchr(field, '=');
  if (p != NULL) {	/* name=value form */
    char *	value;

    value = ++p;
    p	  = strchr(value, '"');
    if (p != NULL) {	/* double quotation */
      value = ++p;
      p = strrchr(value, '"');
      if (p != NULL) {
	*p = EOL;
	return *value == EOL ? NULL : cc_strdup(value);
      }
    }
    return NULL;
  }

  while (*field == SPC) {
    field++;
  }

  return *field == EOL ? NULL : field;
}

/*
 * FUNCTION NAME:	read_entity
 * DESCRIPTION:		read entity of multipart form data
 * ARGUMENTS:		separator, name
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
read_entity(separator, name)
	const char *	separator;
	const char *	name;
{
  size_t	slen = strlen(separator);
  FILE *	fout;
  ListManager *	lmp;
  char		buf[BUFSIZ];
  char		c;
  int		i;
  size_t	size = 0;

  fout = tmpfile();
  if (fout == NULL) {
    cgi_errno = CGIERR_TMPFILE_OPEN_FAILURE;
    return -1;
  }

  lmp = (ListManager *)cc_calloc(1, sizeof(ListManager));
  if (lmp == NULL) {
    fclose(fout);
    return -1;
  }
  lmp->name = cc_strdup(name);
  lmp->fp   = fout;
  lmp->next = files;
  files	    = lmp;

  i = 0;
  while (!feof(stdin)) {
    if (fread(&c, sizeof(char), 1, stdin) > 0) {
      buf[i] = c;
      i++;
      if (i >= BUFSIZ) {
	size += fwrite(buf, sizeof(char), BUFSIZ, fout);
	i = 0;
      }
      else if (c == LF) {
	buf[i]	= EOL;
	if (strncmp(buf, separator, slen) == 0) {
	  (void)fseek(fout, 0L, SEEK_SET);
	  /* remove CRLF in last line */
	  return ftruncate(fileno(fout), (off_t)(size - 2));
	}
	size += fwrite(buf, sizeof(char), (size_t)i, fout);
	i = 0;
      }
    }
  }

  return -1;
}

/*
 * FUNCTION NAME:	get_entity
 * DESCRIPTION:		get entity of multipart form data
 * ARGUMENTS:		separator
 * RETURN VALUE:	Successful: entity, otherwise: NULL
 */
static char *
get_entity(separator)
	const char *	separator;
{
  char		buf[BUFSIZ];
  size_t	slen	= strlen(separator);
  size_t	size	= 0;
  char *	value	= (char *)cc_calloc(1, sizeof(char));

  if (value == NULL) {
    return NULL;
  }
  value[0] = EOL;

  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), stdin) != NULL) {
    size_t	len;

    if (strncmp(buf, separator, slen) == 0) {
      cc_chomp(value);
      return *value == EOL ? NULL : value;
    }
    len	  = strlen(buf);
    size += len;
    value = (char *)cc_realloc(value, size + 1);
    if (value == NULL) {
      cgi_errno = CGIERR_NO_ENOUGH_MEMORY;
      return NULL;
    }
    strlcat(value, buf, size);
  }

  return NULL;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	cm_upload_open
 * DESCRIPTION:		open upload file
 * ARGUMENTS:		upload filename
 * RETURN VALUE:	Successful: file pointer, otherwise: NULL
 */
static FILE *
cm_upload_open(filename)
	const char *	filename;
{
  register ListManager *	lmp;

  for (lmp = files; lmp != NULL; lmp = lmp->next) {
    if (filename != NULL && strcmp(filename, lmp->name) == 0) {
      return lmp->fp;
    }
  }

  return NULL;
}

/*
 * FUNCTION NAME:	cm_upload_store
 * DESCRIPTION:		store upload file
 * ARGUMENTS:		name, filename
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
cm_upload_store(name, filename)
	const char *	name;
	const char *	filename;
{
  register FILE *	fin;
  register FILE *	fout;

  fin = cm_upload_open(name);
  if (fin == NULL) {
    return -1;
  }

  fout = fopen(filename, "w");
  if (fout == NULL) {
    return -1;
  }

  while (!feof(fin)) {
    size_t	len;
    char	buf[BUFSIZ];

    len = fread(buf, sizeof(char), sizeof(buf), fin);
    fwrite(buf, sizeof(char), len, fout);
  }
  fclose(fout);

  (void)fseek(fin, 0L, SEEK_SET);

  return 0;
}

/*
 * FUNCTION NAME:	cm_upload_digest
 * DESCRIPTION:		digest upload file
 * ARGUMENTS:		name, digest_t
 * RETURN VALUE:	Successful: Digest, otherwise: NULL
 */
static char *
cm_upload_digest(fname, type)
	const char *	fname;
	digest_t	type;
{
  register FILE *	fin;

  fin = cm_upload_open(fname);
  if (fin != NULL) {
    size_t	len;
    char	buf[BUFSIZ];

    switch (type) {
      case CC_DIGEST_MD5: {
	  MD5_CTX		md5_ctx;
	  static unsigned char	md5_hex[MD5_DIGEST_LENGTH*2+1];

	  memset(md5_hex, 0, sizeof(md5_hex));
	  MD5Init(&md5_ctx);
	  while ((len = fread(buf, sizeof(char), sizeof(buf), fin)) > 0) {
	    MD5Update(&md5_ctx, buf, len);
	  }
	  MD5End(&md5_ctx, md5_hex);
	  (void)fseek(fin, 0L, SEEK_SET);

	  return md5_hex;
	}
	break;
      case CC_DIGEST_SHA1: {
	  SHA_CTX		sha_ctx;
	  static unsigned char	sha_hex[SHA_DIGEST_LENGTH*2+1];

	  memset(sha_hex, 0, sizeof(sha_hex));
	  SHA1_Init(&sha_ctx);
	  while ((len = fread(buf, sizeof(char), sizeof(buf), fin)) > 0) {
	    SHA1_Update(&sha_ctx, buf, len);
	  }
	  SHA1_End(&sha_ctx, sha_hex);
	  (void)fseek(fin, 0L, SEEK_SET);

	  return sha_hex;
	}
	break;
      case CC_DIGEST_SHA256: {
	  SHA256_CTX		sha256_ctx;
	  static unsigned char	sha256_hex[SHA256_DIGEST_LENGTH*2+1];

	  memset(sha256_hex, 0, sizeof(sha256_hex));
	  SHA256_Init(&sha256_ctx);
	  while ((len = fread(buf, sizeof(char), sizeof(buf), fin)) > 0) {
	    SHA256_Update(&sha256_ctx, buf, len);
	  }
	  SHA256_End(&sha256_ctx, sha256_hex);
	  (void)fseek(fin, 0L, SEEK_SET);

	  return sha256_hex;
	}
	break;
      case CC_DIGEST_RMD160: {
	  RIPEMD160_CTX		rmd160_ctx;
	  static unsigned char	rmd160_hex[RIPEMD160_DIGEST_LENGTH*2+1];

	  memset(rmd160_hex, 0, sizeof(rmd160_hex));
	  RIPEMD160_Init(&rmd160_ctx);
	  while ((len = fread(buf, sizeof(char), sizeof(buf), fin)) > 0) {
	    RIPEMD160_Update(&rmd160_ctx, buf, len);
	  }
	  RIPEMD160_End(&rmd160_ctx, rmd160_hex);
	  (void)fseek(fin, 0L, SEEK_SET);

	  return rmd160_hex;
	}
	break;
      default:
	break;
    }
  }

  return NULL;
}

/*****************************************************************************/

static cmUpload	upload_funcs = {
  cm_upload_open,
  cm_upload_store,
  cm_upload_digest
};

void
cc_upload_register_hooks(void)
{
  cc_hook_module_upload(&upload_funcs);
}
